"""
Convert a quiz into TeX using templates.
"""

import os

import quizcomp.constants
import quizcomp.converter.template

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_TEMPLATE_DIR = os.path.join(THIS_DIR, '..', 'data', 'templates', 'edq-tex')

JINJA_OPTIONS = {
    'block_start_string': '<%',
    'block_end_string': '%>',
    'variable_start_string': '<<',
    'variable_end_string': '>>',
    'comment_start_string': '<#',
    'comment_end_string': '#>',
}

class TexTemplateConverter(quizcomp.converter.template.TemplateConverter):
    def __init__(self, template_dir = DEFAULT_TEMPLATE_DIR,
            cleanup_images = False,
            **kwargs):
        super().__init__(quizcomp.constants.FORMAT_TEX, template_dir,
                cleanup_images = cleanup_images,
                parser_format_options = {
                    'image_path_callback': self._store_images,
                },
                jinja_options = JINJA_OPTIONS, **kwargs)

    def clean_solution_content(self, document):
        tex = document.to_format(quizcomp.constants.FORMAT_TEX)
        if ('\\' not in tex):
            return tex

        content = document.to_format(quizcomp.constants.FORMAT_TEXT)
        content = content.replace('\\', '\\textbackslash{}')

        return content



"""
This file standardizes how we write and read JSON files.
Specifically, we try to be flexible when reading (using JSON5),
and strict when writing (using vanilla JSON).
"""

import json

import json5

def load(file_obj, **kwargs):
    return json5.load(file_obj, **kwargs)

def loads(text, **kwargs):
    return json5.loads(text, **kwargs)

def load_path(path, **kwargs):
    try:
        with open(path, 'r') as file:
            return load(file, **kwargs)
    except Exception as ex:
        raise ValueError(f"Failed to read JSON file '{path}'.") from ex

def dump(data, file_obj, **kwargs):
    return json.dump(data, file_obj, **kwargs)

def dumps(data, **kwargs):
    return json.dumps(data, **kwargs)

def dump_path(data, path, **kwargs):
    with open(path, 'w') as file:
        dump(data, file, **kwargs)



"""
The HTTP session objects here provide a way to make HTTP requests in a way that can be recorded and used later for testing.
There are three types of sessions, all of which respond to get(), post(), and path() methods,
while returning a Response object.

RequestsSession lightly wraps a requests.Session.
This class includes the ability to wait between requests to avoid being rate limited.

SessionRecorder wraps RequestsSession,
and will save the requests and responses to a directory.

TestSession can be loaded from a directory saved by a SessionRecorder,
and will replay any (method, URL) pair that comes to it.
If an unknown request is made, an exception will be raised.
"""

import logging
import os
import re
import time
import urllib.parse

import requests

import quizcomp.util.dirent
import quizcomp.util.hash
import quizcomp.util.json

DEFAULT_WAIT_TIME_SECS = 0.25

# {id: [session, ...], ...}
_test_sessions = {}

def get_session(id, save_http = False, **kwargs):
    """
    Get a session (based on requests.Session) to make HTTP requests.
    If id matches a loaded test session (via load_test_session()), then that session will be returned.
    Then if save_http is true, then a new SessionRecorder will be returned.
    Otherwise return a RequestsSession, which wraps a vanilla requests.Session.
    """

    if (id in _test_sessions):
        session = _test_sessions[id].pop(0)
        if (len(_test_sessions[id]) == 0):
            del _test_sessions[id]

        return session

    if (save_http):
        return SessionRecorder(id, **kwargs)

    return requests.Session(**kwargs)

def load_test_session(id, base_dir):
    """
    Load a test session from a directory into the queue for a specific ID.
    When a test session is used, it is removed from the queue.
    """

    session = TestSession.from_dir(base_dir)

    if (id not in _test_sessions):
        _test_sessions[id] = []

    _test_sessions[id].append(session)

class RequestsSession(object):
    def __init__(self,
            *args,
            wait_time = DEFAULT_WAIT_TIME_SECS,
            **kwargs):
        self._session = requests.Session(**kwargs)

        self._wait_time = wait_time
        self._last_request_time = None

    def get(self, url, *args, **kwargs):
        self._ensure_wait()

        response = self._session.get(url, *args, **kwargs)
        response_data = _response_to_dict(response)
        return Response(response_data)

    def post(self, url, *args, **kwargs):
        self._ensure_wait()

        response = self._session.post(url, *args, **kwargs)
        response_data = _response_to_dict(response)
        return Response(response_data)

    def patch(self, url, *args, **kwargs):
        self._ensure_wait()

        response = self._session.patch(url, *args, **kwargs)
        response_data = _response_to_dict(response)
        return Response(response_data)

    def _ensure_wait(self):
        """
        Ensure that we have waited enough time between requests.
        """

        now = time.time()

        if (self._last_request_time is None):
            self._last_request_time = now
            return

        delta = now - self._last_request_time
        wait_time_secs = self._wait_time - delta

        if (wait_time_secs > 0):
            time.sleep(wait_time_secs)

        self._last_request_time = time.time()

class SessionRecorder(object):
    """
    Record all requests and responses to be later played back via a TestSession.
    """

    def __init__(self, *args, **kwargs):
        self._session = RequestsSession(**kwargs)

        self._count = 0

        self._out_dir = quizcomp.util.dirent.get_temp_path('quizcomp-http-recorded-session-', rm = False)
        logging.info("Saving session HTTP data to '%s'.", self._out_dir)

    def get(self, url, *args, **kwargs):
        return self._wrap_request(url, self._session.get, "GET", *args, **kwargs)

    def post(self, url, *args, **kwargs):
        return self._wrap_request(url, self._session.post, "POST", *args, **kwargs)

    def patch(self, url, *args, **kwargs):
        return self._wrap_request(url, self._session.patch, "PATCH", *args, **kwargs)

    def _wrap_request(self,
            url, request_func, method,
            *args,
            headers = None, params = None, data = None, files = None,
            **kwargs):
        # Remember that we are already wrapping RequestsSession, so this is a Response.
        response = request_func(url, *args,
                headers = headers, params = params, data = data, files = files,
                **kwargs)

        result = {
            'request': {
                'method': method,
                'url': url,
            },
            'response': response._data,
        }

        if ((headers is not None) and (len(headers) > 0)):
            result['request']['headers'] = headers

        if ((params is not None) and (len(params) > 0)):
            result['request']['params'] = params

        if ((data is not None) and (len(data) > 0)):
            result['request']['data'] = data

        if ((files is not None) and (len(files) > 0)):
            result['request']['files'] = list(sorted(files.keys()))

        out_path = self._make_out_path(self._count, method, url, result)
        quizcomp.util.json.dump_path(result, out_path, indent = 4)

        self._count += 1

        return response

    def _make_out_path(self, count, method, url, data):
        method = method.upper()
        clean_filename_part = _clean_filename(url)
        hash_part = _hash_dict(data)

        out_filename = "%03d-%s-%s-%s.json" % (count, method, clean_filename_part, hash_part)
        return os.path.join(self._out_dir, out_filename)

class TestSession(object):
    """
    A session that will only respond with pre-made responses.
    Any request that is not found will raise an exception.
    """

    def __init__(self, responses):
        self._responses = responses

    def get(self, url, *args, **kwargs):
        return self._fetch_response(url, "GET")

    def post(self, url, *args, **kwargs):
        return self._fetch_response(url, "POST")

    def patch(self, url, *args, **kwargs):
        return self._fetch_response(url, "PATCH")

    def _fetch_response(self, url, method):
        key = TestSession._make_key(method, url)
        if (key not in self._responses):
            raise KeyError("Unknown response key: '%s'.", key)

        return Response(self._responses[key]['response'])

    def from_dir(base_dir):
        responses = {}

        for dirent in sorted(os.listdir(base_dir)):
            path = os.path.join(base_dir, dirent)
            data = quizcomp.util.json.load_path(path)

            key = TestSession._make_key(data['request']['method'], data['request']['url'])

            if (key in responses):
                raise ValueError("Duplicate response key '%s' found in '%s'.", key, path)

            responses[key] = data

        return TestSession(responses)

    def _make_key(method, url):
        return "%s::%s" % (method, url)

class Response(object):
    """
    Wrap the data in a requests.Response to provide a consistent interface.
    The data should come from _response_to_dict().
    """

    def __init__(self, data):
        self._data = data

        self.text = data['body']
        self.history = [Response(item) for item in data['history']]

    def raise_for_status(self):
        """
        See https://docs.python-requests.org/en/latest/api/#requests.Response.raise_for_status
        """

        status = self._data['status']
        if (400 <= status < 600):
            raise RuntimeError("HTTP error %d." % (status))

def _clean_filename(url):
    parts = urllib.parse.urlparse(url)
    return parts.hostname + parts.path.replace('/', '_')

def _hash_dict(data):
    json = quizcomp.util.json.dumps(data)
    return quizcomp.util.hash.sha256(json)

def _response_to_dict(response):
    headers = {}
    for (key, value) in response.headers.lower_items():
        headers[key] = value

    history = []
    if (response.history is not None):
        history = [_response_to_dict(item) for item in response.history]

    return {
        'status': response.status_code,
        'headers': headers,
        'body': response.text,
        'history': history,
    }



import os
import sys

import quizcomp.latex
import quizcomp.pdf
import quizcomp.util.dirent

import tests.base

MIN_PDF_SIZE_BYTES = 1000

class PDFConversionTest(tests.base.BaseTest):
    """
    Test PDF generation for quizzes in 'tests/quizzes/good' directory.
    Tests cover both local and Docker-based PDF generation.
    """

    pass

def _add_pdf_tests():
    quiz_files = tests.base.discover_good_quiz_files()
    if (not quiz_files):
        raise ValueError("No quiz.json files found in '%s' or its subdirectories." % (tests.base.GOOD_QUIZZES_DIR))

    for quiz_path in quiz_files:
        _add_pdf_test(quiz_path)

def _add_pdf_test(path):
    base_test_name = os.path.basename(os.path.dirname(path))

    # Test local PDF generation.
    test_name = "test_quiz_pdf_local_%s" % (base_test_name)
    setattr(PDFConversionTest, test_name, _get_quiz_pdf_local_test_method(path))

    # Test Docker PDF generation.
    test_name = "test_quiz_pdf_docker_%s" % (base_test_name)
    setattr(PDFConversionTest, test_name, _get_quiz_pdf_docker_test_method(path))

def _get_quiz_pdf_local_test_method(path):
    """
    Get a test for generating a PDF locally.
    """

    def __method(self):
        _run_pdf_test(self, path, use_docker = False)

    return __method

def _get_quiz_pdf_docker_test_method(path):
    """
    Get a test for generating a PDF with Docker.
    """

    def __method(self):
        _run_pdf_test(self, path, use_docker = True)

    return __method

def _run_pdf_test(self, path, use_docker = False):
    if (use_docker and sys.platform.startswith("win")):
        self.skipTest('Skipping Docker tests on Windows.')

    quizcomp.latex.set_pdflatex_use_docker(use_docker)

    if (not quizcomp.latex.is_available()):
        self.skipTest("`pdflatex` is not available.")

    temp_dir = quizcomp.util.dirent.get_temp_path(prefix = "quizcomp_pdf_test_")
    quiz, variants, _ = quizcomp.pdf.make_with_path(path, base_out_dir = temp_dir)

    for variant in variants:
        pdf_file = os.path.join(temp_dir, quiz.title, f"{variant.title}.pdf")
        self.assertTrue(os.path.exists(pdf_file), f"PDF '{pdf_file}' not generated.")
        self.assertTrue(os.path.getsize(pdf_file) > MIN_PDF_SIZE_BYTES, f"PDF '{pdf_file}' is too small.")

_add_pdf_tests()



"""
Upload quizes to Canvas.
"""

import logging
import os
import urllib.parse
import re

import requests

import quizcomp.common
import quizcomp.constants
import quizcomp.quiz
import quizcomp.util.hash

# TODO(eriq): This code assumes there will never be more than a page of items returned.
PAGE_SIZE = 75

CANVAS_QUIZCOMP_BASEDIR = '/quiz-composer'
CANVAS_QUIZCOMP_QUIZ_DIRNAME = 'quiz'

QUIZ_TYPE_ASSIGNMENT = 'assignment'
QUIZ_TYPE_PRACTICE = 'practice_quiz'

QUESTION_TYPE_MAP = {
    # Direct Mappings
    quizcomp.constants.QUESTION_TYPE_ESSAY: 'essay_question',
    quizcomp.constants.QUESTION_TYPE_FIMB: 'fill_in_multiple_blanks_question',
    quizcomp.constants.QUESTION_TYPE_MATCHING: 'matching_question',
    quizcomp.constants.QUESTION_TYPE_MA: 'multiple_answers_question',
    quizcomp.constants.QUESTION_TYPE_MCQ: 'multiple_choice_question',
    quizcomp.constants.QUESTION_TYPE_MDD: 'multiple_dropdowns_question',
    quizcomp.constants.QUESTION_TYPE_NUMERICAL: 'numerical_question',
    quizcomp.constants.QUESTION_TYPE_TEXT_ONLY: 'text_only_question',
    quizcomp.constants.QUESTION_TYPE_TF: 'true_false_question',
    # Indirect Mappings
    quizcomp.constants.QUESTION_TYPE_FITB: 'short_answer_question',
    quizcomp.constants.QUESTION_TYPE_SA: 'essay_question',
}

QUESTION_FEEDBACK_MAPPING = {
    'general': 'neutral_comments_html',
    'correct': 'correct_comments_html',
    'incorrect': 'incorrect_comments_html',
}

DEFAULT_CANVAS_OPTIONS = {
    'practice': True,
    'published': False,
    'hide_results': None,
    'show_correct_answers': True,
    'allowed_attempts': 1,
    'scoring_policy': 'keep_highest',
    'assignment_group_name': 'Quizzes',
}

ALLOWED_VALUES = {
    'practice': [True, False],
    'published': [True, False],
    'hide_results': [None, 'always', 'until_after_last_attempt'],
    'show_correct_answers': [True, False],
    'scoring_policy': ['keep_highest', 'keep_latest'],
}

class InstanceInfo(object):
    def __init__(self, base_url, course_id, token):
        self.base_url = base_url
        self.course_id = course_id
        self.token = token

        self.context = {}

    def base_headers(self):
        return {
            "Authorization": "Bearer %s" % (self.token),
            "Accept": "application/json+canvas-string-ids",
        }

class CanvasUploader(object):
    def __init__(self, instance, force = False, **kwargs):
        super().__init__(**kwargs)

        if (instance is None):
            raise ValueError("Canvas instance information cannot be None.")

        self.instance = instance
        self.force = force

    def upload_quiz(self, quiz, **kwargs):
        upload_quiz(quiz, self.instance, force = self.force);

def validate_options(old_options):
    options = DEFAULT_CANVAS_OPTIONS.copy()
    options.update(old_options)

    for (key, value) in options.items():
        if (key not in DEFAULT_CANVAS_OPTIONS):
            logging.warning("Unknown canvas options: '%s'." % (key))
            continue

        if (key in ALLOWED_VALUES):
            if (value not in ALLOWED_VALUES[key]):
                raise quizcomp.common.QuizValidationError("Canvas option '%s' has value '%s' not in allowed values: %s." % (key, value, ALLOWED_VALUES[key]))

        if (key == 'allowed_attempts'):
            options['allowed_attempts'] = _validate_allowed_attempts(value)

    return options

def _validate_allowed_attempts(allowed_attempts):
    if (not isinstance(allowed_attempts, (str, int))):
        raise quizcomp.common.QuizValidationError("Allowed attempts must be a positive int (or -1), found '%s'." % (str(allowed_attempts)))

    try:
        allowed_attempts = int(allowed_attempts)
    except:
        raise quizcomp.common.QuizValidationError("Allowed attempts must be a positive int (or -1), found '%s'." % (str(allowed_attempts)))

    if ((allowed_attempts < -1) or (allowed_attempts == 0)):
        raise quizcomp.common.QuizValidationError("Allowed attempts must be a positive int (or -1), found '%s'." % (str(allowed_attempts)))

    return allowed_attempts


def upload_quiz(quiz, instance, force = False):
    """
    Data may be written into the instance context.
    """

    if (not isinstance(quiz, quizcomp.quiz.Quiz)):
        raise ValueError("Canvas quiz uploader requires a quizcomp.quiz.Quiz type, found %s." % (type(quiz)))

    existing_ids = get_matching_quiz_ids(quiz.title, instance)
    if ((len(existing_ids) > 0) and (not force)):
        logging.info("Found a quiz with a matching name '%s', skipping upload.", quiz.title)
        return False

    for existing_id in existing_ids:
        logging.debug("Deleting existing quiz '%s' (%s).", quiz.title, existing_id)
        delete_quiz(existing_id, instance)

    create_quiz(quiz, instance)
    return True

def upload_canvas_files(quiz, instance):
    """
    Canvas requires that images (and other files) be uploaded to their side (instead of embedded),
    so upload all images in one method so we don't upload duplicates.
    """

    # {path: <canvas file id>, ...}
    file_ids = {}

    paths = []
    for group in quiz.groups:
        paths += group.collect_file_paths()

    paths = set(paths)

    for path in paths:
        canvas_path = '/'.join([
            CANVAS_QUIZCOMP_BASEDIR,
            CANVAS_QUIZCOMP_QUIZ_DIRNAME,
            quiz.title,
            quizcomp.util.hash.sha256(path) + os.path.splitext(path)[-1]
        ])

        file_id = upload_file(path, canvas_path, instance)
        file_ids[path] = file_id

    return file_ids

def get_matching_quiz_ids(title, instance):
    response = requests.request(
        method = "GET",
        url = "%s/api/v1/courses/%s/quizzes?per_page=%d" % (instance.base_url, instance.course_id, PAGE_SIZE),
        headers = instance.base_headers())
    response.raise_for_status()

    ids = []
    for quiz in response.json():
        if (quiz['title'] == title):
            ids.append(quiz['id'])

    return ids

def delete_quiz(quiz_id, instance):
    response = requests.request(
        method = "DELETE",
        url = "%s/api/v1/courses/%s/quizzes/%s" % (instance.base_url, instance.course_id, quiz_id),
        headers = instance.base_headers())
    response.raise_for_status()

def fetch_assignment_group(name, instance):
    if (name is None):
        return None

    response = requests.request(
        method = "GET",
        url = "%s/api/v1/courses/%s/assignment_groups?per_page=%d" % (instance.base_url, instance.course_id, PAGE_SIZE),
        headers = instance.base_headers())
    response.raise_for_status()

    for assignment in response.json():
        if (assignment['name'] == name):
            return assignment['id']

    return None

def create_quiz(quiz, instance):
    file_ids = upload_canvas_files(quiz, instance)
    instance.context['file_ids'] = file_ids

    assignment_group_id = fetch_assignment_group(quiz.canvas['assignment_group_name'], instance)

    quiz_type = QUIZ_TYPE_ASSIGNMENT
    if (quiz.canvas['practice']):
        quiz_type = QUIZ_TYPE_PRACTICE

    description = quiz.description.document.to_canvas(canvas_instance = instance, pretty = False)

    data = {
        'quiz[title]': quiz.title,
        'quiz[description]': f"<p>{description}</p><br /><hr /><p>Version: {quiz.version}</p>",
        'quiz[quiz_type]': quiz_type,
        'quiz[published]': quiz.canvas['published'],
        'quiz[assignment_group_id]': assignment_group_id,
        'quiz[time_limit]': quiz.time_limit_mins,
        'quiz[allowed_attempts]': quiz.canvas['allowed_attempts'],
        'quiz[show_correct_answers]': quiz.canvas['show_correct_answers'],
        'quiz[hide_results]': quiz.canvas['hide_results'],
        'quiz[shuffle_answers]': quiz.shuffle_answers,
        'quiz[scoring_policy]': quiz.canvas['scoring_policy'],
    }

    response = requests.request(
        method = "POST",
        url = "%s/api/v1/courses/%s/quizzes" % (instance.base_url, instance.course_id),
        headers = instance.base_headers(),
        data = data)
    response.raise_for_status()

    quiz_id = response.json()['id']

    for question_group in quiz.groups:
        create_question_group(quiz_id, question_group, instance)

def create_question_group(quiz_id, group, instance):
    data = {
        'quiz_groups[][name]': group.name,
        'quiz_groups[][pick_count]': group.pick_count,
        'quiz_groups[][question_points]': group.points,
    }

    response = requests.request(
        method = "POST",
        url = "%s/api/v1/courses/%s/quizzes/%s/groups" % (instance.base_url, instance.course_id, quiz_id),
        headers = instance.base_headers(),
        data = data)
    response.raise_for_status()

    group_id = response.json()['quiz_groups'][0]['id']

    for i in range(len(group.questions)):
        create_question(quiz_id, group_id, group.questions[i], i, instance)

def create_question(quiz_id, group_id, question, index, instance):
    data = _create_question_json(group_id, question, index, instance = instance)

    response = requests.request(
        method = "POST",
        url = "%s/api/v1/courses/%s/quizzes/%s/questions" % (instance.base_url, instance.course_id, quiz_id),
        headers = instance.base_headers(),
        data = data)
    response.raise_for_status()

def _create_question_json(group_id, question, index, instance = None):
    question_type = QUESTION_TYPE_MAP[question.question_type]

    name = question.name
    if (question.custom_header is not None):
        name = question.custom_header

    data = {
        'question[question_type]': question_type,
        'question[question_name]': name,
        'question[quiz_group_id]': group_id,
        # The actual points is taken from the group,
        # but put in a one here so people don't get scared when they see a zero.
        'question[points_possible]': 1,
        'question[position]': index,
        'question[question_text]': question.prompt.document.to_canvas(canvas_instance = instance, pretty = False),
    }

    # Handle question-level feedback.
    for (key, canvas_key) in QUESTION_FEEDBACK_MAPPING.items():
        if (key not in question.feedback):
            continue

        data_key = "question[%s]" % (canvas_key)
        text = question.feedback[key].document.to_canvas(canvas_instance = instance, pretty = False)
        data[data_key] = text

    _serialize_answers(data, question, instance)

    return data

def _serialize_answers(data, question, instance):
    # In Canvas, short answer questions also get mapped to the essay Canvas type.
    if (question.question_type in [quizcomp.constants.QUESTION_TYPE_ESSAY, quizcomp.constants.QUESTION_TYPE_SA]):
        # Essay questions have no answers.
        pass
    elif (question.question_type == quizcomp.constants.QUESTION_TYPE_FIMB):
        _serialize_fimb_answers(data, question, instance)
    elif (question.question_type == quizcomp.constants.QUESTION_TYPE_FITB):
        _serialize_fimb_answers(data, question, instance)
    elif (question.question_type == quizcomp.constants.QUESTION_TYPE_MATCHING):
        _serialize_matching_answers(data, question, instance)
    elif (question.question_type == quizcomp.constants.QUESTION_TYPE_NUMERICAL):
        _serialize_numeric_answers(data, question.answers, instance)
    elif (question.question_type == quizcomp.constants.QUESTION_TYPE_TEXT_ONLY):
        # Text-Only questions have no answers.
        pass
    elif (isinstance(question.answers, list)):
        use_text = (question.question_type == quizcomp.constants.QUESTION_TYPE_TF)
        _serialize_answer_list(data, question.answers, instance, use_text = use_text)
    elif (isinstance(question.answers, dict)):
        count = 0
        for key, item in question.answers.items():
            _serialize_answer_list(data, item['values'], instance,
                    start_index = count, blank_id = key, use_text = True)
            count += len(item['values'])
    else:
        raise ValueError(f"Unknown answers type '{type(question.answers)}'.")

def _serialize_answer_list(data, answers, instance,
        start_index = 0, blank_id = None, use_text = False):
    for i in range(len(answers)):
        _serialize_answer(data, answers[i], start_index + i, instance,
            blank_id = blank_id, use_text = use_text)

def _serialize_answer(data, answer, index, instance, blank_id = None, use_text = False):
    weight = 0
    if (answer.is_correct()):
        weight = 100

    data["question[answers][%d][answer_weight]" % (index)] = weight

    if (use_text):
        text = answer.document.to_text(text_allow_special_text = True, text_allow_all_characters = True)
        data["question[answers][%d][answer_text]" % (index)] = text
    else:
        html = answer.document.to_canvas(canvas_instance = instance, pretty = False)
        data["question[answers][%d][answer_html]" % (index)] = html

    if (blank_id is not None):
        data["question[answers][%d][blank_id]" % (index)] = blank_id

    if (answer.has_feedback()):
        feedback_html = answer.feedback.document.to_canvas(canvas_instance = instance, pretty = False)
        data["question[answers][%d][answer_comment_html]" % (index)] = feedback_html

def _serialize_matching_answers(data, question, instance):
    for i in range(len(question.answers['matches'])):
        left_content = question.answers['matches'][i]['left'].document.to_text(text_allow_special_text = True, text_allow_all_characters = True)
        right_content = question.answers['matches'][i]['right'].document.to_text(text_allow_special_text = True, text_allow_all_characters = True)

        data["question[answers][%d][answer_match_left]" % (i)] = left_content
        data["question[answers][%d][answer_match_right]" % (i)] = right_content

        if (question.answers['matches'][i]['left'].has_feedback()):
            text = question.answers['matches'][i]['left'].feedback.document.to_canvas(canvas_instance = instance, pretty = False)
            data["question[answers][%d][answer_comment_html]" % (i)] = text

    if (len(question.answers['distractors']) > 0):
        distractors = [distractor.document.to_text(text_allow_special_text = True, text_allow_all_characters = True) for distractor in question.answers['distractors']]
        data["question[matching_answer_incorrect_matches]"] = "\n".join(distractors)

def _serialize_fimb_answers(data, question, instance):
    index = 0

    for (key, item) in question.answers.items():
        key_text = item['key'].document.to_text()

        for i in range(len(item['values'])):
            value_text = item['values'][i].document.to_text(text_allow_special_text = True, text_allow_all_characters = True)

            data[f"question[answers][{index}][blank_id]"] = key_text
            data[f"question[answers][{index}][answer_weight]"] = 100
            data[f"question[answers][{index}][answer_text]"] = value_text

            if (item['values'][i].has_feedback()):
                feedback_text = item['values'][i].feedback.document.to_canvas(canvas_instance = instance, pretty = False)
                data[f"question[answers][{index}][answer_comment_html]"] = feedback_text

            index += 1

def _serialize_numeric_answers(data, answers, instance):
    # Note that the keys/constants for numerical answers are different than what the documentation says:
    # https://canvas.instructure.com/doc/api/quiz_questions.html#QuizQuestion

    for i in range(len(answers)):
        answer = answers[i]

        data[f"question[answers][{i}][answer_weight]"] = 100
        data[f"question[answers][{i}][numerical_answer_type]"] = answer.type + '_answer'

        if (answer.type == quizcomp.constants.NUMERICAL_ANSWER_TYPE_EXACT):
            data[f"question[answers][{i}][answer_exact]"] = answer.value
            data[f"question[answers][{i}][answer_error_margin]"] = answer.margin
        elif (answer.type == quizcomp.constants.NUMERICAL_ANSWER_TYPE_RANGE):
            data[f"question[answers][{i}][answer_range_start]"] = answer.min
            data[f"question[answers][{i}][answer_range_end]"] = answer.max
        elif (answer.type == quizcomp.constants.NUMERICAL_ANSWER_TYPE_PRECISION):
            data[f"question[answers][{i}][answer_approximate]"] = answer.value
            data[f"question[answers][{i}][answer_precision]"] = answer.precision
        else:
            raise ValueError(f"Unknown numerical answer type: '{answer.type}'.")

        if (answer.has_feedback()):
            feedback_text = answer.feedback.document.to_canvas(canvas_instance = instance, pretty = False)
            data[f"question[answers][{i}][answer_comment_html]"] = feedback_text

def upload_file(path, canvas_path, instance):
    parent_id = ensure_folder(os.path.dirname(canvas_path), instance)
    upload_url, upload_params = _init_file_upload(path, canvas_path, parent_id, instance)
    file_id = _upload_file_contents(path, upload_url, upload_params)

    return file_id

def _init_file_upload(path, canvas_path, parent_id, instance):
    canvas_name = os.path.basename(canvas_path)

    size = os.stat(path).st_size

    data = {
        'name': canvas_name,
        'size': size,
        'parent_folder_id': parent_id,
        'on_duplicate': 'overwrite',
    }

    response = requests.request(
        method = "POST",
        url = "%s/api/v1/courses/%s/files" % (instance.base_url, instance.course_id),
        headers = instance.base_headers(),
        data = data)
    response.raise_for_status()

    response = response.json()

    upload_url = response['upload_url']
    upload_params = response['upload_params']

    return upload_url, upload_params

def _upload_file_contents(path, upload_url, upload_params):
    files = {
        'file': open(path, 'rb'),
    }

    response = requests.request(
        method = "POST",
        url = upload_url,
        data = upload_params,
        files = files)
    response.raise_for_status()

    location = response.headers.get('Location', None)
    if (location is None):
        raise ValueError(f"Could not find location for uploaded file: '{path}'.")

    file_id = os.path.basename(urllib.parse.urlparse(location).path)

    return file_id

def ensure_folder(canvas_path, instance):
    folder_id = get_folder(canvas_path, instance)
    if (folder_id is not None):
        return folder_id

    folder_id = create_folder(canvas_path, instance)

    # Canvas will not hide created parents.
    hide_folder(CANVAS_QUIZCOMP_BASEDIR, instance)

    return folder_id

def get_folder(canvas_path, instance):
    # The canvas path should be absolute.
    response = requests.request(
        method = "GET",
        url = "%s/api/v1/courses/%s/folders/by_path%s" % (instance.base_url, instance.course_id, canvas_path),
        headers = instance.base_headers())

    if (response.status_code == 404):
        return None

    response.raise_for_status()

    return response.json()[-1]['id']

def create_folder(canvas_path, instance):
    name = os.path.basename(canvas_path)
    parent_path = os.path.dirname(canvas_path)

    data = {
        'name': name,
        'parent_folder_path': parent_path,
        # Canvas wants a string here despite the documentation saying it is a bool.
        'hidden': 'true',
    }

    response = requests.request(
        method = "POST",
        url = "%s/api/v1/courses/%s/folders" % (instance.base_url, instance.course_id),
        headers = instance.base_headers(),
        data = data)
    response.raise_for_status()

    folder_id = response.json()['id']

    return folder_id

def hide_folder(canvas_path, instance):
    folder_id = get_folder(canvas_path, instance)
    return hide_folder_id(folder_id, instance)

def hide_folder_id(folder_id, instance):
    data = {
        # Canvas wants a string here despite the documentation saying it is a bool.
        # TODO(eriq): Make a bug request?
        'hidden': 'true',
    }

    response = requests.request(
        method = "PUT",
        url = "%s/api/v1/folders/%s" % (instance.base_url, folder_id),
        headers = instance.base_headers(),
        data = data)
    response.raise_for_status()



import atexit
import os
import shutil
import tempfile
import uuid

def read_file_if_exists(path, **kwargs):
    """
    Read the file if it exists, otherwise return None.
    """

    if (os.path.exists(path)):
        return read_file(path, **kwargs)

    return None

def read_file(path, strip = True, rstrip = True):
    with open(path, 'r') as file:
        contents = file.read()

    if (strip):
        contents = contents.strip()

    if (rstrip):
        contents = contents.rstrip()

    return contents

def write_file(path, contents):
    with open(path, 'w') as file:
        file.write(contents)

def get_temp_path(prefix = '', suffix = '', rm = True, mkdir = True):
    """
    Get a path to a valid temp dirent.
    mkdir will determine if this directory will exist on return.
    If rm is True, then the dirent will be attempted to be deleted on exit
    (no error will occur if the path is not there).
    """

    path = None
    while ((path is None) or os.path.exists(path)):
        path = os.path.join(tempfile.gettempdir(), prefix + str(uuid.uuid4()) + suffix)

    if (mkdir):
        os.makedirs(path, exist_ok = True)

    if (rm):
        atexit.register(remove_dirent, path)

    return path

def remove_dirent(path):
    if (not os.path.exists(path)):
        return

    if (os.path.isfile(path) or os.path.islink(path)):
        os.remove(path)
    elif (os.path.isdir(path)):
        shutil.rmtree(path)
    else:
        raise ValueError("Unknown type of dirent: '%s'." % (path))

def copy_dirent(source, dest, **kwargs):
    """
    Copy a file or directory into dest.
    kwargs will be passed along to the underlying function:
    shutil.copy2 for files and shutil.copytree for dirs.
    See each specific function for semantics.
    """

    if (os.path.isfile(source)):
        shutil.copy2(source, dest, **kwargs)
    else:
        shutil.copytree(source, dest, **kwargs)



import os
import re

import quizcomp.constants
import quizcomp.converter.convert
import quizcomp.quiz
import quizcomp.util.json
import tests.base

EXPECTED_FILENAME = 'expected.json'

class TestTemplateConversion(tests.base.BaseTest):
    """
    Test template conversion.
    """

    def _assert_exists_replace(self, container, key, replacement):
        """
        Ensure a value exists (and is not None), and then replace it.
        The old value will be returned.
        """

        value = container.get(key, None)
        self.assertIsNotNone(value, f"Key '{key}' does not exist.")

        container[key] = replacement

        return value

def _add_good_convert_questions():
    for quiz_path in tests.base.discover_good_quiz_files():
        test_name = _make_name('good_convert', quiz_path)
        setattr(TestTemplateConversion, test_name, _get_good_convert_test(quiz_path))

def _get_good_convert_test(quiz_path):
    """
    That that the tests here rely on the `quizcomp.converter.json.JSONTemplateConverter` converter.
    This is not meant for production use and may not have full functionality.
    The main difficulty it has is when converting answers to an easy format for testing.
    As more tests are developed, this conversion may need to be adjusted to better test different aspects,
    e.g., feedback is currently ignored when converting answers.
    """

    def __method(self):
        base_dir = os.path.dirname(quiz_path)
        name = os.path.basename(base_dir)

        expected_path = os.path.join(base_dir, EXPECTED_FILENAME)
        if (not os.path.exists(expected_path)):
            self.fail(f"Expected quiz output does not exist: '{expected_path}'.")

        expected = quizcomp.util.json.load_path(expected_path)

        quizcomp.converter.convert

        quiz = quizcomp.quiz.Quiz.from_path(quiz_path)
        variant = quiz.create_variant()
        raw_result = quizcomp.converter.convert.convert_variant(variant, format = quizcomp.constants.FORMAT_JSON_TEMPLATE)

        result = quizcomp.util.json.loads(raw_result)

        # Clean up the result.
        self._assert_exists_replace(result['quiz'], 'seed', 0)
        self._assert_exists_replace(result['quiz'], 'version', "test")

        # Clean up question base dirs specially by making them relative to the tests directory.
        for group in result['groups']:
            for question in group['questions']:
                base_dir = self._assert_exists_replace(question, 'base_dir', '')
                rel_dir = os.path.relpath(base_dir, tests.base.TESTS_DIR)
                question['base_dir'] = rel_dir

        # Convert the paths in the expected output to the system path separator.
        for group in expected.get('groups', []):
            for question in group['questions']:
                question['base_dir'] = os.path.join(*question['base_dir'].split('/'))

        self.assertJSONDictEqual(expected, result)

    return __method

def _add_bad_validate_questions():
    for quiz_path in tests.base.discover_bad_quiz_files():
        test_name = _make_name('bad_validate', quiz_path)
        setattr(TestTemplateConversion, test_name, _get_bad_validate_test(quiz_path))

def _get_bad_validate_test(quiz_path):
    def __method(self):
        try:
            quiz = quizcomp.quiz.Quiz.from_path(quiz_path)
        except Exception:
            # Expected.
            return

        self.fail("Failed to raise an exception.")

    return __method

def _make_name(prefix, path):
    dirname = os.path.basename(os.path.dirname(path))
    dirname = _clean_name_part(dirname)

    return "test_%s__%s" % (prefix, dirname)

def _clean_name_part(text):
    clean_text = text.lower().strip().replace(' ', '_')
    clean_text = re.sub(r'\W+', '', clean_text)
    return clean_text

def _apply_text_options(options, a, b):
    if (options.get("ignore-whitespace", False)):
        a = re.sub(r'\s+', '', a)
        b = re.sub(r'\s+', '', b)

    return a, b

_add_bad_validate_questions()
_add_good_convert_questions()



import copy
import types

ENCODING = 'utf-8'

BASE_DIR_KEY = 'base_dir'

CONTEXT_ENV_KEY = 'qg_context'
CONTEXT_KEY_STYLE = 'style'
CONTEXT_KEY_IMAGE_CALLBACK = 'image_path_callback'
CONTEXT_KEY_FORCE_RAW_IMAGE_SRC = 'force_raw_image_src'
CONTEXT_KEY_TEXT_ALLOW_ALL_CHARACTERS = 'text_allow_all_characters'
CONTEXT_KEY_TEXT_ALLOW_SPECIAL_TEXT = 'text_allow_special_text'

TOKEN_META_KEY_ROOT = 'qg_root'
TOKEN_META_KEY_STYLE = 'qg_style'

CONTENT_NODES = {
    'code_block',
    'code_inline',
    'fence',
    'math_block',
    'math_inline',
    'placeholder',
    'text',
    'text_special',
}

def handle_block_style(block_info, context):
    """
    Handle style associated with a block by extracting the style, prepping a context with that style,
    and returning (context, full style, block style).
    """

    block_style = block_info.get(TOKEN_META_KEY_STYLE, {})
    style = context.get(CONTEXT_KEY_STYLE, {})

    if (len(block_style) > 0):
        style.update(block_style)
        context_value = {CONTEXT_KEY_STYLE: style}
        context = prep_context(context, context_value)

    return context, style, block_style

def prep_context(context, options = {}):
    """
    Return an immutable copy of the context with any of the specified additional value.
    Will make a deep copy if necessary.
    """

    if (isinstance(context, types.MappingProxyType) and (len(options) == 0)):
        return context

    context = dict(context)
    if (len(options) > 0):
        context = _partial_deep_copy(context)
        context.update(options)

    return types.MappingProxyType(context)

def _partial_deep_copy(source_dict):
    """
    Only deep copy values that are dicts or lists, shallow copy the rest.
    This is especially important for types that are fully or mostly imutable or callables.
    """

    result = {}
    for (key, value) in source_dict.items():
        if (isinstance(value, (dict, list))):
            value = copy.deepcopy(value)

        result[key] = value

    return result



import logging
import os
import shutil
import subprocess

import quizcomp.util.encoding

_node_bin_dir = None

def set_node_bin_dir(path):
    global _node_bin_dir
    _node_bin_dir = path

def _has_command(command, cwd = '.'):
    result = subprocess.run(["which", command], cwd = cwd, capture_output = True)
    return (result.returncode == 0)

def _has_package(package, cwd = '.'):
    bin_path = 'npm'
    if (_node_bin_dir is not None):
        bin_path = os.path.join(_node_bin_dir, bin_path)

    result = subprocess.run([bin_path, "list", package], cwd = cwd, capture_output = True)
    return (result.returncode == 0)

def is_available(cwd = '.'):
    if ((_node_bin_dir is None) and (shutil.which('npx') is None)):
        logging.warning("Could not find `npx` (usually installed with `npm`), cannot use katex equations.")
        return False

    if (not _has_package('katex', cwd = cwd)):
        logging.warning("Could not find the `katex` NodeJS package, cannot use katex equations.")
        return False

    return True

def to_html(text, cwd = '.'):
    bin_path = 'npx'
    if (_node_bin_dir is not None):
        bin_path = os.path.join(_node_bin_dir, bin_path)

    result = subprocess.run([bin_path, "katex", "--format", "mathml"], cwd = cwd,
        input = text, encoding = quizcomp.util.encoding.DEFAULT_ENCODING,
        capture_output = True)

    if (result.returncode != 0):
        raise ValueError("KaTeX did not exit cleanly. Stdout: '%s', Stderr: '%s'" % (result.stdout, result.stderr))

    return result.stdout

def set_cli_args(parser):
    parser.add_argument('--nodejs-bin-dir', dest = 'node_bin_dir',
        action = 'store', type = str, default = None,
        help = ('A NodeJS binary directory that includes `npm` and `npx`.'
                + ' If not specified, $PATH will be searched.'
                + ' Used for HTML equations.'))

    return parser

def init_from_args(args):
    if (args.node_bin_dir is not None):
        set_node_bin_dir(args.node_bin_dir)

    return args



import quizcomp.common
import quizcomp.constants
import quizcomp.parser.public
import quizcomp.question.base

class MDD(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_MDD):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        self._check_type(self.answers, dict, "'answers' key")

        if (len(self.answers) == 0):
            raise quizcomp.common.QuestionValidationError(self, "No answers provided, at least one answer required.")

        for (key, answers) in list(self.answers.items()):
            # If this was already in the full MDD format, then we need to pull out the values.
            if ((isinstance(answers, dict)) and ('values' in answers)):
                answers = answers['values']

            values = self._validate_answer_list(answers, self.base_dir, min_correct = 1, max_correct = 1)

            self.answers[key] = {
                'key': quizcomp.parser.public.parse_text(key, base_dir = self.base_dir),
                'values': values,
            }

        self._check_placeholders(self.answers.keys())

    def _shuffle(self, rng):
        for key in self.answers:
            rng.shuffle(self.answers[key]['values'])



import glob
import os

import quizcomp.common
import quizcomp.constants
import quizcomp.question.base
import quizcomp.quiz
import quizcomp.util.dirent
import quizcomp.util.serial

class Project(quizcomp.util.serial.JSONSerializer):
    """
    A project object represents a directory that may contain quizzes and questions.
    The directory (or any file structure) is never serialized.
    """

    def __init__(self, type = quizcomp.constants.TYPE_PROJECT,
            name = '',
            base_dir = '',
            **kwargs):
        super().__init__(type = type, **kwargs)

        self.name = name
        self._base_dir = base_dir

        try:
            self.validate(cls = Project, **kwargs)
        except Exception as ex:
            raise quizcomp.common.QuizValidationError("Error while validating project '%s' ('%s')." % (self._base_dir, self.base_dir)) from ex

    def _validate(self, **kwargs):
        if ((self._base_dir is None) or (self._base_dir == "")):
            raise quizcomp.common.QuizValidationError("Base directory cannot be empty.")

        if (not os.path.isdir(self._base_dir)):
            raise quizcomp.common.QuizValidationError("Base directory '%s' does not exist or is not a directory." % (self._base_dir))

    def find_resources(self):
        """
        Find all the resources associated with this project.
        Returns the path to these resources as (quizzes, questions).
        """

        quizzes = sorted(glob.glob(os.path.join(self._base_dir, '**', quizcomp.constants.QUIZ_FILENAME), recursive = True))
        questions = sorted(glob.glob(os.path.join(self._base_dir, '**', quizcomp.constants.QUESTION_FILENAME), recursive = True))

        return (quizzes, questions)

    def load_resources(self, **kwargs):
        """
        Load (and validate) all the resources associated with this project.
        Returns: ([(quiz path, quiz object), ...], [(question path, question object), ...])
        """

        quiz_paths, question_paths = self.find_resources()

        quizzes = [(path, quizcomp.quiz.Quiz.from_path(path, **kwargs)) for path in quiz_paths]
        questions = [(path, quizcomp.question.base.Question.from_path(path, **kwargs)) for path in question_paths]

        return (quizzes, questions)

    def save(self, out_dir = None):
        """
        Save this project to the specified dir,
        or our own base dir if no other dir is specified.

        Note that in the case we are saving to our own base dir,
        this operation just (re)writes the project file.
        """

        if (out_dir is None):
            out_dir = self._base_dir
        out_dir = os.path.abspath(out_dir)

        if (os.path.isfile(out_dir)):
            raise ValueError("Output for project save cannot be a file, it must be a directory.")

        # If we are pointing to the base_dir, then just save the project file and return.
        if (os.path.isfile(out_dir) and os.path.samefile(out_dir, self._base_dir)):
            self.to_path(os.path.join(self._base_dir, quizcomp.constants.PROJECT_FILENAME))
            return

        quizcomp.util.dirent.copy_dirent(self._base_dir, out_dir,
                symlinks = True, dirs_exist_ok = True)
        self.to_path(os.path.join(out_dir, quizcomp.constants.PROJECT_FILENAME))

    @classmethod
    def from_path(cls, path, **kwargs):
        # If we are looking at a dir, assume the project file is directly inside.
        if (os.path.isdir(path)):
            path = os.path.join(path, quizcomp.constants.PROJECT_FILENAME)

        path = os.path.abspath(path)

        # If the project file does not exist, then continue with minimal info.
        if (not os.path.exists(path)):
            return Project(base_dir = os.path.dirname(path), **kwargs)

        return super().from_path(path, **kwargs)



"""
The `quizcomp.cli.parse` package contains tools for parsing various objects/files from disk.
"""

import sys

import quizcomp.util.cli

def main():
    return quizcomp.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())



import glob
import logging
import os

import quizcomp.common
import quizcomp.constants
import quizcomp.question.base
import quizcomp.util.serial

class Group(quizcomp.util.serial.JSONSerializer):
    def __init__(self, name = '',
            pick_count = 1, points = 10,
            shuffle_answers = True, pick_with_replacement = True,
            custom_header = None, skip_numbering = None,
            hints = None, hints_first = None, hints_last = None,
            questions = [],
            ids = {},
            **kwargs):
        super().__init__(**kwargs)

        self.name = name
        self.pick_count = pick_count
        self.points = points

        self.shuffle_answers = shuffle_answers
        self.pick_with_replacement = pick_with_replacement
        self._used_question_indexes = set()

        self.custom_header = custom_header
        self.skip_numbering = skip_numbering

        self.hints = hints
        self.hints_first = hints_first
        self.hints_last = hints_last

        self.questions = questions

        try:
            self.validate()
        except Exception as ex:
            ids = ids.copy()
            ids[name] = self.name

            raise quizcomp.common.QuizValidationError('Error while validating group.', ids = ids) from ex

    def _validate(self, **kwargs):
        if ((self.name is None) or (self.name == "")):
            raise quizcomp.common.QuizValidationError("Name cannot be empty.")

        if (self.pick_count < 0):
            raise quizcomp.common.QuizValidationError("Pick count cannot be negative.")

        if (self.hints is None):
            self.hints = {}

        if (self.hints_first is None):
            self.hints_first = {}

        if (self.hints_last is None):
            self.hints_last = {}

        if (not isinstance(self.questions, list)):
            raise quizcomp.common.QuizValidationError("Questions must be a non-empty list, found: '%s'." % (str(self.questions)))

        if (len(self.questions) == 0):
            raise quizcomp.common.QuizValidationError("Questions must be non-empty.")

        for question in self.questions:
            question.inherit_from_group(self)

        if (self.pick_count > len(self.questions)):
            logging.warning("Group '%s' was asked to pick more questions than available (pick_count: %d, group size: %d)." % (
                    self.name, self.pick_count, len(self.questions)))
            self.pick_count = len(self.questions)

    def collect_file_paths(self):
        paths = []

        for question in self.questions:
            paths += question.collect_file_paths()

        return paths

    @staticmethod
    def from_dict(group_info, base_dir, ids = {}, **kwargs):
        group_info = group_info.copy()

        paths = []
        for path in group_info.get('questions', []):
            if (not os.path.isabs(path)):
                path = os.path.join(base_dir, path)
            paths.append(os.path.abspath(path))

        paths = list(sorted(set(paths)))

        questions = []
        for path in paths:
            questions += _parse_questions(path)

        group_info['questions'] = questions

        return Group(**group_info, ids = ids)

    def choose_questions(self, all_questions = False, rng = None, with_replacement = True):
        if ((self.pick_count == 0) or (len(self.questions) == 0)):
            logging.warning("Group '%s' will select no questions (pick_count: %d, group size: %d)." % (
                    self.name, self.pick_count, len(self.questions)))
            return []

        with_replacement = (self.pick_with_replacement and with_replacement)

        if (rng is None):
            seed = random.randint(0, 2**64)
            rng = random.Random(seed)

        count = self.pick_count
        if (all_questions):
            count = len(self.questions)

        questions = self._choose_questions(count, rng, with_replacement)

        # Rename questions if there are more than one.
        if (len(questions) > 1):
            for i in range(len(questions)):
                questions[i].name = "%s - %d" % (self.name, i + 1)

        # Inherit position-specific hints.
        questions[0].add_hints(self.hints_first)
        questions[-1].add_hints(self.hints_last)

        return questions

    def _choose_questions(self, count, rng, with_replacement):
        indexes = list(range(len(self.questions)))

        if (not with_replacement):
            indexes = list(set(indexes) - self._used_question_indexes)

            if (count > len(indexes)):
                logging.warning("Group '%s' does not have enough questions to pick without replacement." % (self.name))
                # Reset the selection pool.
                indexes = list(range(len(self.questions)))
                self._used_question_indexes = set()

        rng.shuffle(indexes)
        indexes = indexes[:count]

        if (not with_replacement):
            self._used_question_indexes |= set(indexes)

        return [self.questions[index].copy() for index in indexes]

def _parse_questions(path):
    if (not os.path.exists(path)):
        raise quizcomp.common.QuizValidationError(f"Question path does not exist: '{path}'.")

    if (os.path.isfile(path)):
        return [quizcomp.question.base.Question.from_path(path)]

    questions = []
    for subpath in sorted(glob.glob(os.path.join(path, '**', quizcomp.constants.QUESTION_FILENAME), recursive = True)):
        questions.append(quizcomp.question.base.Question.from_path(subpath))

    return questions



import mdformat.plugins
import mdformat.renderer

import quizcomp.parser.common
import quizcomp.parser.math

def math_inline(node, context):
    qg_context = context.env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
    return quizcomp.parser.math._render_md(node.content, True, qg_context)

def math_block(node, context):
    qg_context = context.env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
    return quizcomp.parser.math._render_md(node.content, False, qg_context)

def placeholder(node, context):
    return "<placeholder>%s</placeholder>" % (node.content)

def container_block(node, context):
    # We can ignore blocks when outputting markdown (since it is non-standard).
    # Just render the child node (there should only be one).
    if ((node.children is None) or (len(node.children) == 0)):
        return ''

    parts = []
    for child in node.children:
        parts.append(child.render(context))

    return "\n\n".join(parts)

class QuizComposerMDformatExtension(mdformat.plugins.ParserExtensionInterface):
    CHANGES_AST = False
    POSTPROCESSORS = {}
    RENDERERS = {
        'container_block': container_block,
        'math_block': math_block,
        'math_inline': math_inline,
        'placeholder': placeholder,
    }

def get_renderer(options):
    extensions = options.get('parser_extension', [])
    extensions += [
        mdformat.plugins.PARSER_EXTENSIONS['tables'],
        QuizComposerMDformatExtension(),
    ]
    options['parser_extension'] = extensions

    return mdformat.renderer.MDRenderer(), options



"""
Pieces of the parsing infrastructure that are exposed for public use.
Code outside this package should generally only use these resources.
"""

import os

import quizcomp.parser.parse
import quizcomp.util.serial
import quizcomp.util.dirent

class ParsedText(quizcomp.util.serial.PODSerializer):
    """
    A representation of text that has been successfully parsed.
    """

    def __init__(self, text, document):
        self.text = text
        self.document = document

    def to_pod(self, **kwargs):
        return self.text

def parse_text(text, base_dir = '.'):
    text, document = quizcomp.parser.parse._parse_text(text, base_dir)
    return ParsedText(text, document)

def parse_file(path):
    if (not os.path.isfile(path)):
        raise ValueError(f"Path to parse ('{path}') is not a file.")

    text = quizcomp.util.dirent.read_file(path)
    base_dir = os.path.dirname(path)

    return parse_text(text, base_dir = base_dir)



import quizcomp.common
import quizcomp.quiz

DUMMY_QUIZ_DATA = {
    'title': 'Dummy Title',
    'description': 'Dummy description.',
    'course_title': 'Dummy Course',
    'term_title': 'Dummy Term',
    'version': '0.0.0',
}

DUMMY_GROUP_DATA = {
    'name': 'Dummy Question',
}

class Variant(quizcomp.quiz.Quiz):
    """
    A quiz varint is an instantiation of a quiz with specific set of questions chosen for each group.
    Variants still have question groups, but each group must only have the exact number of questions required for each group
    (or it is a validation error).

    Variants created directly from quizzes (as opposed to from a JSON file)
    will already have all the correct components, and will therefore only be lightly validated.
    Quizzes created from files will undergo full validation.
    """

    def __init__(self, type = quizcomp.constants.TYPE_VARIANT,
            **kwargs):
        super().__init__(type = type, **kwargs)
        self.validate(cls = Variant, **kwargs)

        self.questions = []
        for group in self.groups:
            self.questions += group.questions

    def _validate(self, **kwargs):
        # Ensure that each group has the correct number of questions.
        for i in range(len(self.groups)):
            group = self.groups[i]

            if (len(group.questions) != group.pick_count):
                raise quizcomp.common.QuizValidationError("Group at index %d (%s) has %d questions, expecting exactly %d." % (i, group.name, len(group.questions), group.pick_count))

    @staticmethod
    def get_dummy(question):
        """
        Get a "dummy" variant that has no real information.
        """

        quiz_data = DUMMY_QUIZ_DATA.copy()
        group_data = DUMMY_GROUP_DATA.copy()

        group_data['questions'] = [question]
        group_data['_skip_class_validations'] = [quizcomp.group.Group]
        quiz_data['groups'] = [quizcomp.group.Group(**group_data)]

        return Variant(**quiz_data)



import re

import markdown_it
import mdit_py_plugins.container
import mdit_py_plugins.dollarmath

import quizcomp.parser.document
import quizcomp.util.json

_parser = None
_options = None

EXTRA_OPTIONS = [
    'table',
]

PLUGINS = [
    (mdit_py_plugins.dollarmath.dollarmath_plugin, {}),
    (mdit_py_plugins.container.container_plugin, {'name': 'block'}),
]

HTML_TOKENS = {
    'html_block',
    'html_inline',
}

def _get_parser():
    _parser = markdown_it.MarkdownIt('commonmark')

    for option in EXTRA_OPTIONS:
        _parser.enable(option)

    for (plugin, options) in PLUGINS:
        _parser.use(plugin, **options)

    _options = _parser.options

    return _parser, _options

def _clean_text(text):
    # Remove carriage returns.
    text = text.replace("\r", '')

    # Trim whitespace.
    text = text.strip();

    return text

# Returns (transformed text, tokens).
def _parse_text(text, base_dir):
    text = _clean_text(text)

    parser, _ = _get_parser()

    tokens = parser.parse(text)
    tokens = _post_process(tokens)

    document = quizcomp.parser.document.ParsedDocument(tokens, base_dir = base_dir)

    return (text.strip(), document)

def _post_process(tokens):
    """
    Post-process the token stream.
    This allows us to edit the AST without needing the change the parser.
    """

    tokens = _add_root_block(tokens)
    tokens = _process_placeholders(tokens)
    tokens = _process_style(tokens)
    tokens = _process_html(tokens)
    tokens = _remove_empty_tokens(tokens)

    return tokens

def _add_root_block(tokens):
    """
    Add a root block element to the document.
    """

    if (len(tokens) == 0):
        return []

    open_token = markdown_it.token.Token('container_block_open', 'div', 1)
    open_token.block = True
    open_token.map = list(tokens[0].map)
    open_token.attrJoin('class', 'qg-root-block')
    open_token.meta[quizcomp.parser.common.TOKEN_META_KEY_ROOT] = True

    close_token = markdown_it.token.Token('container_block_close', 'div', -1)

    return [open_token] + tokens + [close_token]

def _process_style(tokens, containing_block = None):
    """
    Locate any style nodes, parse them, remove them, and hoist their content to the containing block.
    """

    remove_indexes = []
    for i in range(len(tokens)):
        token = tokens[i]

        # If this is a block, then mark it as the current block.
        # Any discovered style get's hoisted to the containing block.
        if (token.type == 'container_block_open'):
            containing_block = token
        elif ((token.type in HTML_TOKENS) and (token.content.strip().startswith('<style>'))):
            # Style nodes are HTML with a 'style' tag.

            if (containing_block is None):
                raise ValueError("Found a style node that does not have a containing block.")

            style = _process_style_content(token.content)
            containing_block.meta[quizcomp.parser.common.TOKEN_META_KEY_STYLE] = style

            # Mark this token for removal.
            remove_indexes.append(i)

        # Check all children.
        if (_has_children(token)):
            token.children = _process_style(token.children, containing_block = containing_block)

    for remove_index in sorted(list(set(remove_indexes)), reverse = True):
        tokens.pop(remove_index)

    return tokens

def _process_style_content(raw_content):
    raw_content = raw_content.strip()

    # Get content without tags ('<style>', '</style>').
    content = re.sub(r'\s+', ' ', raw_content)
    content = re.sub(r'^<style>(.*)</style>$', r'\1', content).strip()

    # Ignore empty style.
    if (len(content) == 0):
        return {}

    # If the content does not start with a '{', then assume the braces were left out and add them.
    # We will also ignore content that starts with a '[' (a JSON list), that will be handled later.
    if (content[0] not in ['{', '[']):
        content = "{%s}" % (content)

    try:
        style = quizcomp.util.json.loads(content)
        if (not isinstance(style, dict)):
            raise ValueError("Style is not a JSON object, found: '%s'." % (type(style)))
    except Exception as ex:
        raise ValueError(('Failed to load style tag.'
                + ' Style content must be a JSON object (start/end braces may be omitted).'
                + " Original exception message: '%s'." % (ex)
                + " Found:\n---\n%s\n---" % (raw_content)))

    return style

def _remove_empty_tokens(tokens):
    """
    Remove any inline's without text or blocks without children.
    """

    # Keep looping until nothing is removed.
    while True:
        remove_indexes = []
        for i in range(len(tokens)):
            token = tokens[i]

            # Remove empty leaf content nodes.
            if ((token.type in quizcomp.parser.common.CONTENT_NODES) and (token.content == '')):
                remove_indexes.append(i)

            # Check children for removal.
            if (_has_children(token)):
                token.children = _remove_empty_tokens(token.children)

                # Remove nodes that have been emptied out.
                if (len(token.children) == 0):
                    remove_indexes.append(i)

            # Check for empty containers.
            # Look for this token being the open, and the next token being the close.
            if ((i < (len(tokens) - 1)) and (token.type.endswith('_open')) and (tokens[i + 1].type.endswith('_close'))):
                next_token = tokens[i + 1]
                base_type = re.sub('_open$', '', token.type)
                next_base_type = re.sub('_close$', '', next_token.type)

                # Remove if types match and neither has any kids (close should never have kids).
                if ((base_type == next_base_type) and (not _has_children(token)) and (not _has_children(next_token))):
                    remove_indexes += [i, i + 1]

        for remove_index in sorted(list(set(remove_indexes)), reverse = True):
            tokens.pop(remove_index)

        if (len(remove_indexes) == 0):
            break

    return tokens

def _process_placeholders(tokens):
    """
    Find any placeholder HTML tags and replace them with a placeholder token.
    plceholder tags must either be an HTML block or inline with the same parent.
    """

    remove_indexes = []

    # Use a non-standard loop so that we can manually advance the index within the loop.
    i = -1
    while (i < (len(tokens) - 1)):
        i += 1
        token = tokens[i]

        if (token.type == 'html_block'):
            if ((not token.content) or (not token.content.strip().startswith('<placeholder'))):
                continue

            # Replace the HTML block token with a placeholder token.
            tokens[i] = _create_placeholder_token(token)
        elif (token.type == 'html_inline'):
            if ((not token.content) or (not token.content.strip().startswith('<placeholder'))):
                continue

            open_tag_index = i
            close_tag_index = None

            # Look for the close tag at this same level (under the same parent).
            for j in range(i + 1, len(tokens)):
                other_token = tokens[j]
                if (other_token.type != 'html_inline'):
                    continue

                if ((not other_token.content) or (not other_token.content.strip() == '</placeholder>')):
                    continue

                close_tag_index = j
                break

            if (close_tag_index is None):
                raise ValueError("Could not find closing tag for <placeholder>.")

            if ((close_tag_index - open_tag_index) < 2):
                raise ValueError("Did not find any content inside a <placeholder> tag.")

            if ((close_tag_index - open_tag_index) > 2):
                raise ValueError("Found too much content inside a <placeholder> tag, it shoud have only plain text.")

            text_token_index = open_tag_index + 1
            text_token = tokens[text_token_index]
            if (text_token.type != 'text'):
                raise ValueError("Found non-text content inside a <placeholder> tag, it shoud have only plain text.")

            # All tokens (open, content/label, close) are accounted for.
            # Replace the content node and remove the open/close tags.
            tokens[text_token_index] = _create_placeholder_token(text_token)
            remove_indexes += [open_tag_index, close_tag_index]

            # Advance to the close token.
            i = close_tag_index

        if (_has_children(token)):
            token.children = _process_placeholders(token.children)

    for remove_index in sorted(list(set(remove_indexes)), reverse = True):
        tokens.pop(remove_index)

    return tokens

def _create_placeholder_token(token):
    # Fetch the label in the tag.
    label = re.sub(r'\s+', ' ', token.content.strip())
    label = re.sub(r'^<placeholder.*>(.*)</placeholder>$', r'\1', label).strip()
    if (len(label) == 0):
        raise ValueError("Found an empty '<placeholder>' tag.")

    return markdown_it.token.Token(
            type = 'placeholder', tag = '', nesting = 0,
            map = token.map, content = label)

def _process_html(tokens):
    """
    Remove or replacec all HTML tags.
    This will recursively descend to find all tags.
    Line breaks will be replaced with hard breaks.
    """

    remove_indexes = []
    for i in range(len(tokens)):
        token = tokens[i]

        if (token.type in HTML_TOKENS):
            if (token.content.strip().startswith('<br')):
                tokens[i] = markdown_it.token.Token(
                        type = 'hardbreak', tag = 'br', nesting = 0,
                        map = token.map)
            else:
                remove_indexes.append(i)

        if (_has_children(token)):
            token.children = _process_html(token.children)

    for remove_index in sorted(list(set(remove_indexes)), reverse = True):
        tokens.pop(remove_index)

    return tokens

def _has_children(token):
    return ((token.children is not None) and (len(token.children) > 0))



import logging
import os
import shutil
import subprocess

_pdflatex_bin_path = None
_pdflatex_use_docker = False

DOCKER_IMAGE = "ghcr.io/edulinq/pdflatex-docker:1.0.0"

def set_pdflatex_bin_path(path):
    global _pdflatex_bin_path
    _pdflatex_bin_path = path

def set_pdflatex_use_docker(pdflatex_use_docker):
    global _pdflatex_use_docker
    _pdflatex_use_docker = pdflatex_use_docker

def is_available():
    if (_pdflatex_use_docker):
        if (not _is_docker_available()):
            logging.warning("Docker is not available, cannot compile PDFs.")
            return False

        return True

    if (_pdflatex_bin_path is not None):
        return True

    if (shutil.which('pdflatex') is None):
        logging.warning("Could not find `pdxlatex`, cannot compile PDFs.")
        return False

    return True

def _is_docker_available():
    if (shutil.which('docker') is None):
        return False

    result = subprocess.run(["docker", "info"], stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL)
    return (result.returncode == 0)

def compile(path):
    """
    Compile a LaTeX file to PDF in its containing directory.

    The caller must provide a path to a TeX file within a prepared output directory.
    This directory should contain all necessary resources (e.g., images) and no non-relevant files.
    Compilation may generate additional files (e.g., .aux, .log) in this directory,
    and permissions may be modified as needed.
    """

    if (_pdflatex_use_docker is True):
        _compile_docker(path)
    else:
        _compile_local(path)

def _compile_local(path):
    bin_path = "pdflatex"
    if (_pdflatex_bin_path is not None):
        bin_path = _pdflatex_bin_path

    tex_filename = os.path.basename(path)
    out_dir = os.path.dirname(path)

    # Need to compile twice to get positioning information.
    for _ in range(2):
        result = subprocess.run([bin_path, '-interaction=nonstopmode', tex_filename],
                                cwd = out_dir, capture_output = True)
        if (result.returncode != 0):
            raise ValueError("pdflatex did not exit cleanly. Stdout: '%s', Stderr: '%s'" % (result.stdout, result.stderr))

def _compile_docker(path):
    tex_filename = os.path.basename(path)
    out_dir_path = os.path.abspath(os.path.dirname(path))

    docker_cmd = [
        "docker", "run", "--rm",
        "-v", f"{out_dir_path}:/work",
        DOCKER_IMAGE,
        tex_filename
    ]

    result = subprocess.run(docker_cmd, capture_output = True, text = True)
    if (result.returncode != 0):
        raise ValueError("Docker compilation failed with exit code '%s'. Stdout: '%s', Stderr: '%s'" % (result.returncode, result.stdout, result.stderr))

def set_cli_args(parser):
    parser.add_argument('--pdflatex-bin-path', dest = 'pdflatex_bin_path',
        action = 'store', type = str, default = None,
        help = ('The path to the pdflatex binary to use.'
                + ' If not specified, $PATH will be searched.'
                + ' Used to compile PDFs.'))

    parser.add_argument('--pdflatex-use-docker', dest = 'pdflatex_use_docker',
        action = 'store_true', default = False,
        help = ('Use Docker to compile PDFs with pdflatex.'
                + " The Docker image '%s' will be used." % (DOCKER_IMAGE)))

    return parser

def init_from_args(args):
    if (args.pdflatex_use_docker):
        set_pdflatex_use_docker(args.pdflatex_use_docker)

    if (args.pdflatex_bin_path is not None):
        set_pdflatex_bin_path(args.pdflatex_bin_path)

    return args



"""
Create and upload a GradeScope PDF quiz.
"""

import logging
import os
import sys

import quizcomp.args
import quizcomp.pdf
import quizcomp.uploader.gradescope
import quizcomp.util.dirent
import quizcomp.util.json

def run(args):
    quiz, variants, options = quizcomp.pdf.make_with_args(args, write_options = False)
    out_dir = options['out_dir']

    options['gradescope'] = {
        'course': args.course_id,
        'rubric': args.rubric,
        'user': args.user,
    }

    gradescope_ids = []

    for i in range(len(variants)):
        variant = variants[i]

        uploader = quizcomp.uploader.gradescope.GradeScopeUploader(args.course_id, args.user, args.password,
                force = args.force, rubric = args.rubric,
                save_http = args.save_http)
        gradescope_id, created = uploader.upload_quiz(variant, base_dir = out_dir)

        options['variants'][i]['gradescope_id'] = gradescope_id

    # If there are multiple variants and all variants were created (non were skipped).
    if (len(variants) > 1):
        ids = [variants_data['gradescope_id'] for variants_data in options['variants']]

        uploader = quizcomp.uploader.gradescope.GradeScopeUploader(args.course_id, args.user, args.password,
                save_http = args.save_http)
        uploader.create_assignment_group(quiz.title, ids)
        logging.info("Created GradeScope Assignment Group: '%s'.", quiz.title)

    path = os.path.join(out_dir, quizcomp.pdf.OPTIONS_FILENAME)
    with open(path, 'w') as file:
        quizcomp.util.json.dump(options, file, indent = 4)

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(
            prog = 'quizcomp.cli.gradescope.upload',
            description = "Create and upload a GradeScope PDF quiz.")

    quizcomp.pdf.set_cli_args(parser)

    parser.add_argument('--rubric', dest = 'rubric',
        action = 'store_true', default = False,
        help = 'Create an initial rubric for this assignment in GradeScope (default: %(default)s).')

    parser.add_argument('--course', dest = 'course_id',
        action = 'store', type = str, required = True,
        help = 'Course ID to upload the quiz under.')

    parser.add_argument('--user', dest = 'user',
        action = 'store', type = str, required = True,
        help = 'The user to authenticate as.')

    parser.add_argument('--pass', dest = 'password',
        action = 'store', type = str, required = True,
        help = 'The password to use.')

    parser.add_argument('--force', dest = 'force',
        action = 'store_true', default = False,
        help = 'Override (delete) any exiting quiz with the same name (default: %(default)s).')

    parser.add_argument('--save-http', dest = 'save_http',
        action = 'store_true', default = False,
        help = 'Save any http requests to a debugging directory (default: %(default)s).')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import quizcomp.constants
import quizcomp.question.base

class MA(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_MA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        self._validate_self_answer_list()

    def _shuffle(self, rng):
        self._shuffle_answers_list(rng)



import logging
import os

import quizcomp.parser.common
import quizcomp.parser.image
import quizcomp.parser.renderer.base
import quizcomp.parser.style

TEX_REPLACEMENTS = {
    # Specially handle braces and slashes to avoid clobbering other replacements.
    '{': 'ZZZzzz  OPEN BRACE REPLACEMENT  zzzZZZ',
    '}': 'ZZZzzz  CLOSE BRACE REPLACEMENT  zzzZZZ',
    '\\': 'ZZZzzz  BACKSLASH REPLACEMENT  zzzZZZ',

    '|': '\\textbar{}',
    '$': '\\$',
    '#': '\\#',
    '%': '\\%',
    '^': '\\^',
    '_': '\\_',
    'π': '$\\pi$',
    '`': '\\`{}',

    'ZZZzzz  OPEN BRACE REPLACEMENT  zzzZZZ': '\\{',
    'ZZZzzz  CLOSE BRACE REPLACEMENT  zzzZZZ': '\\}',
    'ZZZzzz  BACKSLASH REPLACEMENT  zzzZZZ': '\\textbackslash{}',
}

VERB_CHARACTERS = ['|', '!', '@', '#', '$', '^', '&', '-', '_', '=', '+']

TEX_TEXT_TABLE_ALIGNMENT = {
    quizcomp.parser.style.ALLOWED_VALUES_ALIGNMENT_LEFT: 'l',
    quizcomp.parser.style.ALLOWED_VALUES_ALIGNMENT_CENTER: 'c',
    quizcomp.parser.style.ALLOWED_VALUES_ALIGNMENT_RIGHT: 'r',
}

# HTML and CommonMark support six levels of headings,
# we will use these analogues for TeX.
# The only standard TeX heading we are not using is \part{},
# which is before \chapter{}.
HEADINGS = [
    'chapter',
    'section',
    'subsection',
    'subsubsection',
    'paragraph',
    'subparagraph',
]

class QuizComposerRendererTex(quizcomp.parser.renderer.base.QuizComposerRendererBase):
    def _container_block(self, node, context):
        # Pull any style attatched to this block and put it in a copy of the context.
        context, full_style, block_style = quizcomp.parser.common.handle_block_style(node, context)

        # Compute fixes using different styles depending on if this block is root.
        # If we are root, then we need to use all style.
        # If we are not root, then earlier blocks would have already applied other style,
        # and we only need the block style.
        active_style = block_style
        if (node.get(quizcomp.parser.common.TOKEN_META_KEY_ROOT, False)):
            active_style = full_style

        prefixes, suffixes = quizcomp.parser.style.compute_tex_fixes(active_style)
        child_content = [self._render_node(child, context) for child in node.children()]

        content = prefixes + child_content + list(reversed(suffixes))

        return "\n\n".join(content)

    def _text(self, node, context):
        return tex_escape(node.text())

    def _softbreak(self, node, context):
        return "\n"

    def _hardbreak(self, node, context):
        return '~\\newline\n'

    def _em(self, node, context):
        content = ''.join([self._render_node(child, context) for child in node.children()])
        return r"\textit{%s}" % (content)

    def _strong(self, node, context):
        content = ''.join([self._render_node(child, context) for child in node.children()])
        return r"\textbf{%s}" % (content)

    def _fence(self, node, context):
        language_string = ''

        info = node.get('info', None)
        if ((info is not None) and (len(info) > 0)):
            language_string = "[language=%s]" % (info)

        return "\\begin{lstlisting}%s\n%s\n\\end{lstlisting}" % (language_string, node.text().rstrip())

    def _code_block(self, node, context):
        """
        This token is poorly named, it is actually an indented code block.
        Treat it like a fence with no info string.
        """

        return self._fence(node, context)

    def _code_inline(self, node, context):
        text = node.text()

        delim = None
        for char in VERB_CHARACTERS:
            if (char not in text):
                delim = char
                break

        if (delim is None):
            raise ValueError("Could not find a delimiter to use with TeX's `\verb'.")

        return r"\verb%s%s%s" % (delim, text, delim)

    def _math_block(self, node, context):
        text = node.text().strip()
        return f"$$\n{text}\n$$"

    def _math_inline(self, node, context):
        text = node.text().strip()
        return f"$ {text} $"

    def _image(self, node, context):
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})
        base_dir = context.get(quizcomp.parser.common.BASE_DIR_KEY, '.')
        callback = context.get(quizcomp.parser.common.CONTEXT_KEY_IMAGE_CALLBACK, None)

        src = node.get('src', '')
        src = quizcomp.parser.image.handle_callback(callback, src, base_dir)

        width_float = quizcomp.parser.style.get_image_width(style)
        path = os.path.realpath(os.path.join(base_dir, src))

        return r"\includegraphics[width=%0.2f\textwidth]{%s}" % (width_float, src)

    def _link(self, node, context):
        text = ''.join([self._render_node(child, context) for child in node.children()]).strip()
        url = node.get('href', '')

        if (len(text) == 0):
            return r"\url{%s}" % (url)

        return r"\href{%s}{%s}" % (url, text)

    def _placeholder(self, node, context):
        text = tex_escape(node.text())
        return r"\textsc{<%s>}" % (text)

    def _table(self, node, context):
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})

        border_table = quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_BORDER_TABLE, quizcomp.parser.style.DEFAULT_TABLE_BORDER_TABLE)
        border_cells = quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_BORDER_CELLS, quizcomp.parser.style.DEFAULT_TABLE_BORDER_CELLS)
        default_alignment = quizcomp.parser.style.get_alignment(style, quizcomp.parser.style.KEY_TEXT_ALIGN, default_value = quizcomp.parser.style.ALLOWED_VALUES_ALIGNMENT_CENTER)

        column_infos = _discover_column_info(node)

        column_join = ''
        if (border_cells):
            column_join = '|'

        # Build column specifiers.
        column_specifiers = []
        for column_info in column_infos:
            raw_alignment = column_info.get('text-align', default_alignment)
            column_specifiers.append(TEX_TEXT_TABLE_ALIGNMENT[raw_alignment])
        column_spec = column_join.join(column_specifiers)

        if (border_table):
            column_spec = '|' + column_spec + '|'

        lines = [
            r'\begin{tabular}{ ' + column_spec + ' }',
        ]

        if (border_table):
            lines.append(r'\hline')

        lines.append("\n".join([self._render_node(child, context) for child in node.children()]))

        if (border_table):
            lines.append(r'\hline')

        lines += [
            r'\end{tabular}',
        ]

        lines = _apply_tex_table_style(style, lines)

        return "\n".join(lines)

    def _thead(self, node, context):
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})
        head_rule = quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_HEAD_RULE, quizcomp.parser.style.DEFAULT_TABLE_HEAD_RULE)

        content = "\n".join([self._render_node(child, context) for child in node.children()])

        if (head_rule):
            content += '\n\\hline'

        return content

    def _tbody(self, node, context):
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})
        border_cells = quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_BORDER_CELLS, quizcomp.parser.style.DEFAULT_TABLE_BORDER_CELLS)

        delim = "\n"
        if (border_cells):
            delim = "\n\\hline\n"

        content = delim.join([self._render_node(child, context) for child in node.children()])
        return content

    def _tr(self, node, context):
        content = ' & '.join([self._render_node(child, context) for child in node.children()])
        return content + ' \\\\'

    def _th(self, node, context):
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})
        bold = quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_HEAD_BOLD, quizcomp.parser.style.DEFAULT_TABLE_HEAD_BOLD)

        content = ''.join([self._render_node(child, context) for child in node.children()])

        if (bold):
            content = "\\textbf{%s}" % (content)

        return content

    def _td(self, node, context):
        content = ''.join([self._render_node(child, context) for child in node.children()])
        return content

    def _bullet_list(self, node, context):
        items = '\n'.join([self._render_node(child, context) for child in node.children()])
        return "\\begin{itemize}\n" + items + "\n\\end{itemize}"

    def _ordered_list(self, node, context):
        items = '\n'.join([self._render_node(child, context) for child in node.children()])
        return "\\begin{enumerate}\n" + items + "\n\\end{enumerate}"

    def _list_item(self, node, context):
        content = ''.join([self._render_node(child, context) for child in node.children()])
        return "    \\item " + content

    def _hr(self, node, context):
        return "\\hrulefill"

    def _heading(self, node, context):
        level = self.parse_heading_level(node)

        index = level - 1
        if ((index < 0) or (index >= len(HEADINGS))):
            raise ValueError("Heading index is out of range: %d." % (index))

        heading = HEADINGS[index]
        content = ''.join([self._render_node(child, context) for child in node.children()])

        return "\\%s{%s}" % (heading, content)

    def _blockquote(self, node, context):
        content = ''.join([self._render_node(child, context) for child in node.children()])
        return "\\begin{quote}\n%s\n\\end{quote}" % (content)

def get_renderer(options):
    return QuizComposerRendererTex(), options

def tex_escape(text):
    """
    Prepare normal text for tex.
    """

    for key, value in TEX_REPLACEMENTS.items():
        text = text.replace(key, value)

    return text

def _apply_tex_table_style(style, lines):
    # Height
    height = max(1.0, float(style.get(quizcomp.parser.style.KEY_TABLE_CELL_HEIGHT, quizcomp.parser.style.DEFAULT_TABLE_CELL_HEIGHT)))
    prefix = [
        r'\begingroup',
        r'\renewcommand{\arraystretch}{%0.2f}' % (height),
    ]
    lines = prefix + lines
    lines.append(r'\endgroup')

    # Width
    width = max(1.0, float(style.get(quizcomp.parser.style.KEY_TABLE_CELL_WIDTH, quizcomp.parser.style.DEFAULT_TABLE_CELL_WIDTH)))
    prefix = [
        r'\begingroup',
        r'\setlength{\tabcolsep}{%0.2fem}' % (width),
    ]
    lines = prefix + lines
    lines.append(r'\endgroup')

    return lines

def _discover_column_info(node, info = None):
    """
    Descend into a table to figure out how many columns there are and if a column has alignment information.
    We will not make assumptions about symmetric rows or existence of headers.
    The largest row will determine width, and the last instance of styling information in a column will determine alignment.
    """

    if (info is None):
        info = []

    type = node.type()
    if (type not in {'table', 'thead', 'tbody', 'tr', 'td', 'th'}):
        return info

    children = node.children()

    if (type in {'table', 'thead', 'tbody'}):
        for child in children:
            info = _discover_column_info(child, info = info)
        return info

    if (type == 'tr'):
        for i in range(len(children)):
            if (i >= len(info)):
                info.append({})

            info[i] = _discover_column_info_cells(children[i], style = info[i])

        return info

    logging.warning("Unexpected table node type '%s'." % (type))
    return info

def _discover_column_info_cells(node, style = {}):
    type = node.type()
    if (type not in {'td', 'th'}):
        return style

    raw_style = node.get('style', '').strip()
    if (len(raw_style) == 0):
        return style

    rules = [part.strip() for part in raw_style.split(';')]
    for raw_rule in rules:
        key, value = [part.strip() for part in raw_rule.split(':', maxsplit = 1)]
        style[key] = value

    return style



"""
A place to handle common CLI arguments.
"parsers" in this file are always assumed to be argparse parsers.

The general idea is that callers can register callbacks to be called before and after parsing CLI arguments.
Pre-callbacks are generally intended to add arguments to the parser,
while post-callbacks are generally intended to act on the results of parsing.

Common args are registered at the bottom.
"""

import argparse

# Import modules that have common args to ensure they are loaded.
import quizcomp.katex
import quizcomp.latex
import quizcomp.log

# {module name: function(parser), ...}
_pre = {}

# {module name: function(args), ...}
_post = {}

class Parser(argparse.ArgumentParser):
    """
    Extend an argparse parser to call the pre and post functions.
    """

    def parse_args(self, *args, skip_modules = [], **kwargs):
        pre_parse(self, skip_modules)
        args = super().parse_args(*args, **kwargs)
        post_parse(args, skip_modules)

        return args

def pre_parse(parser, skip_modules = []):
    return _run(parser, _pre, skip_modules)

def post_parse(args, skip_modules = []):
    return _run(args, _post, skip_modules)

def _run(payload, functions, skip_modules):
    for (module_name, callback) in functions.items():
        if (module_name in skip_modules):
            continue

        callback(payload)

    return payload

# Register args under a module name.
def register(module_name, pre_callback = None, post_callback = None):
    if (pre_callback is not None):
        _pre[module_name] = pre_callback

    if (post_callback is not None):
        _post[module_name] = post_callback

# Register common args.
register('log', quizcomp.log.set_cli_args, quizcomp.log.init_from_args)
register('katex', quizcomp.katex.set_cli_args, quizcomp.katex.init_from_args)
register('latex', quizcomp.latex.set_cli_args, quizcomp.latex.init_from_args)



import quizcomp.constants
import quizcomp.question.base

class Essay(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_ESSAY):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        self._validate_text_answers()



import os

import quizcomp.util.dirent
import quizcomp.util.git
import tests.base

class TestGit(tests.base.BaseTest):
    def test_in_repo(self):
        # This test should live inside a repo.
        version = quizcomp.util.git.get_version(tests.base.TESTS_DIR)
        self.assertNotEqual(quizcomp.util.git.UNKNOWN_VERSION, version, 'Got an unknown version. (This can also happen if this instance of the quizcomp project is not in a git repo.')

    def test_cwd(self):
        # The tests should be run in a repo (this project repo).
        version = quizcomp.util.git.get_version()
        self.assertNotEqual(quizcomp.util.git.UNKNOWN_VERSION, version, 'Got an unknown version. (This can also happen if the tests are not being run inside of a git repo.')

    def test_not_in_repo(self):
        # A new temp dir should not be in a git repo.
        path = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp-test-git-')
        version = quizcomp.util.git.get_version(path)
        self.assertEqual(quizcomp.util.git.UNKNOWN_VERSION, version)



import re

import lxml.etree

import quizcomp.parser.parse
import quizcomp.parser.renderer.canvas
import quizcomp.parser.renderer.html
import quizcomp.parser.renderer.markdown
import quizcomp.parser.renderer.tex
import quizcomp.parser.renderer.text

def canvas(tokens, env = {}, pretty = False, **kwargs):
    _, options = quizcomp.parser.parse._get_parser()

    renderer, options = quizcomp.parser.renderer.canvas.get_renderer(options)
    raw_html = renderer.render(tokens, options, env)

    return clean_html(raw_html, pretty = pretty)

def html(tokens, env = {}, pretty = False, **kwargs):
    _, options = quizcomp.parser.parse._get_parser()

    renderer, options = quizcomp.parser.renderer.html.get_renderer(options)
    raw_html = renderer.render(tokens, options, env)

    return clean_html(raw_html, pretty = pretty)

def md(tokens, env = {}, **kwargs):
    _, options = quizcomp.parser.parse._get_parser()

    renderer, options = quizcomp.parser.renderer.markdown.get_renderer(options)
    content = renderer.render(tokens, options, env)

    return content.strip()

def tex(tokens, env = {}, **kwargs):
    _, options = quizcomp.parser.parse._get_parser()

    renderer, options = quizcomp.parser.renderer.tex.get_renderer(options)
    content = renderer.render(tokens, options, env)

    return content.strip()

def text(tokens, env = {}, **kwargs):
    _, options = quizcomp.parser.parse._get_parser()

    renderer, options = quizcomp.parser.renderer.text.get_renderer(options)
    content = renderer.render(tokens, options, env)

    return content.strip()

def render(format, tokens, env = {}, **kwargs):
    """
    General routing render function.
    """

    render_function = globals().get(format, None)
    if (render_function is None):
        raise ValueError("Could not find render function: 'quizcomp.parser.render.%s'." % (format))

    return render_function(tokens, env = env, **kwargs)

def clean_html(raw_html, pretty = False):
    """
    Clean up and standardize the HTML.
    If |pretty|, then the output will be indented properly, and extra space will be stripped (which may mess with some inline spacing).
    |pretty| should only be used when being read by a human for visual inspection.
    """

    raw_html = raw_html.strip()
    if (len(raw_html) == 0):
        return raw_html

    parser = lxml.etree.XMLParser(remove_blank_text = True)
    root = lxml.etree.fromstring(raw_html, parser)
    content = lxml.etree.tostring(root, pretty_print = pretty, encoding = 'unicode')

    return content.strip()



import quizcomp.constants
import quizcomp.converter.html
import quizcomp.converter.json
import quizcomp.converter.tex
import quizcomp.converter.qti
import quizcomp.variant

SUPPORTED_FORMATS = [
    quizcomp.constants.FORMAT_CANVAS,
    quizcomp.constants.FORMAT_HTML,
    quizcomp.constants.FORMAT_JSON,
    quizcomp.constants.FORMAT_TEX,
    quizcomp.constants.FORMAT_QTI,
]

# Formats for testing only.
TEST_SUPPORTED_FORMAT = [
    quizcomp.constants.FORMAT_JSON_TEMPLATE,
]

def get_converter_class(format = quizcomp.constants.FORMAT_JSON):
    if (format == quizcomp.constants.FORMAT_JSON):
        return quizcomp.converter.json.JSONConverter
    elif (format == quizcomp.constants.FORMAT_HTML):
        return quizcomp.converter.html.HTMLTemplateConverter
    elif (format == quizcomp.constants.FORMAT_CANVAS):
        return quizcomp.converter.html.CanvasTemplateConverter
    elif (format == quizcomp.constants.FORMAT_TEX):
        return quizcomp.converter.tex.TexTemplateConverter
    elif (format == quizcomp.constants.FORMAT_QTI):
        return quizcomp.converter.qti.QTITemplateConverter
    elif (format == quizcomp.constants.FORMAT_JSON_TEMPLATE):
        return quizcomp.converter.json.JSONTemplateConverter
    else:
        raise ValueError("No known converter for format '%s'." % (format))

def get_converter(format = quizcomp.constants.FORMAT_JSON, **kwargs):
    converter_class = get_converter_class(format = format)
    return converter_class(**kwargs)

def convert_variant(variant, format = quizcomp.constants.FORMAT_JSON,
        constructor_args = {}, converter_args = {}):
    if (not isinstance(variant, quizcomp.variant.Variant)):
        raise ValueError("convert_variant() requires a quizcomp.variant.Variant type, found %s." % (type(variant)))

    converter = get_converter(format = format, **constructor_args)
    return converter.convert_variant(variant, **converter_args)

def convert_question(question, format = quizcomp.constants.FORMAT_JSON,
        constructor_args = {}, converter_args = {}):
    converter = get_converter(format = format, **constructor_args)
    return converter.convert_question(question, **converter_args)
